% Implementation of an FHRR Vector Symbolic Architecture
% See: Tony Plate. Distributed Representations and
% Nested Compositional Structure. PhD thesis, Toronto, Canada, 1994.
classdef FHRR < handle
    properties
        d % number of dimensions, e.g. 4096
    end
    
    methods
        % d is number of dimensions, e.g. 4096
        function obj = FHRR(d)
            obj.d = d;
        end
        
        function R = getRandomVecs(obj, n)
            R = pi* (2 * (0.5-rand(n, obj.d)));
        end
        
        % Two operation modes:
        % - two inputs: A, B can be matrices, corresponding rows are bound
        % - single input: all rows of A are bound
        function R = bind(obj, A, B)
            if exist('B', 'var')
                R = A+B;
            else
                R = sum(A);
            end
            R = wrapToPi(R);
        end
        
        % All rows of A are bundled
        % w is weight factors, one for each row of A
        function [R, ASS, ACS] = bundle(obj, A, w)
    
            AC = cos(A);
            AS = sin(A);
            
            if exist('w', 'var')
                % weighted bundle
                n = size(A,1);
                W = repmat(reshape(w,n,1), 1, obj.d); % expand to a weight Matrix
                ACS = sum(W.*AC,1);
                ASS = sum(W.*AS,1); 
            else
                ACS = sum(AC,1);
                ASS = sum(AS,1); 
            end
            
            R = atan2(ASS, ACS);
        end
        
        % Similar to bundle() but already accepty sin(A) and cos(A) as
        % input instead of internally computing
        % All rows of A (given by AC=cos(A) and AS=sin(A)) are bundled
        % w is weight factors, one for each row of A
        function R = bundleComplex(obj, AS, AC, w)
            
            if exist('w', 'var')
                % weighted bundle
                n = size(AS,1);
                W = repmat(reshape(w,n,1), 1, obj.d); % expand to a weight Matrix
                ACS = sum(W.*AC,1);
                ASS = sum(W.*AS,1); 
            else
                ACS = sum(AC,1);
                ASS = sum(AS,1); 
            end
            
            R = atan2(ASS, ACS);
        end
        
        % use base vector B to encode scalar p with scale alpha 
        function R = encode(obj, B, p, alpha)
            R = wrapToPi(B .* (p*alpha));
        end
        
        function S = computeSim(obj, v1, v2) 
            % convert to complex values
            v1_c = exp(v1*1i);
            v2_c = exp(-v2*1i);

            S= real(conj(v1_c)*v2_c')/size(v1,2);
        end
        
    end
end