% Use the FHRR Vector Symbolic Architecture to combine all information from 
% all entities in a single descriptor vector.
% 
% E ... set of entities (output of getEntities())
% Y ... list of local appearance features
%
% EE ... the output descriptor
%
function EE = encodeEntities(E,Y,params)

    vsa = FHRR(4096);

    nc = numel(params.relevantClasses);
    
    % configure some internal parameters
    nx=5; ny=5;
    nxFB=4; nyFB=6; % the #intervalls 
    
    % initialize some random vectors
    rng(0); v.gradDir = vsa.getRandomVecs(8);
    rng(1); v.classLabels = vsa.getRandomVecs(nc);
    rng(2); v.grid = vsa.getRandomVecs(nx*ny);
    v.gridSin = sin(v.grid);
    v.gridCos = cos(v.grid);
    
    v.histCellEnc = zeros(nc*8, vsa.d);
    idx=0;
    for ori = 1:8 % this ordering is crucial to be compatible with weights = gradStat(:);
        for b=1:nc
            idx=idx+1;
            v.histCellEnc(idx, :) = vsa.bind( [v.classLabels(b,:); v.gradDir(ori,:)] );
        end
    end
    v.sinHistCellEnc = sin(v.histCellEnc);
    v.cosHistCellEnc = cos(v.histCellEnc);
    rng(100); v.baseX = vsa.getRandomVecs(1);
    rng(101); v.baseY = vsa.getRandomVecs(1);
    
    
    % preprocess appearance features
    AF = preprocessAppearanceFeatures(Y, params);        
    
    % process each entity
    encE = zeros(numel(E),vsa.d);
    encEsin = zeros(numel(E),vsa.d);
    encEcos = zeros(numel(E),vsa.d);
    entity_weights = zeros(numel(E),1);    
    for i=1:numel(E)
       
        
        % encode class
        C = v.classLabels(E(i).Class, :);
        
        % encode neighborhood relations
        N = encodeGradStat(vsa, E(i).GradStat, v);
        
        % encode pose  
        S = shapeFromSimpleGridEncoding(vsa, E(i).PixelList, v, nx, ny, E(i).ImageSize(1), E(i).ImageSize(2));
               
        % use binding to obtain a single vector for the spatial semantic information
        SSI = vsa.bind([C;S;N]);
        
        % encode appearance based on local features on an entity
        [F, nKP, LKP] = encodeLocalFeatures(vsa, Y.keypoints, [0 0], AF, E(i).PixelList, v, nxFB, nyFB, E(i).ImageSize(1), E(i).ImageSize(2));
        
        % use bundling to combine SSI and appearance F        
        if nKP>0
            [encE(i,:), encEsin(i,:), encEcos(i,:)] = vsa.bundle( [SSI; F] ); 
        else
            [encE(i,:), encEsin(i,:), encEcos(i,:)] = vsa.bundle( SSI );   
        end
            
        % compute the weight of the entity based on its size
        entity_weights(i) = sqrt(E(i).Area);   
        
    end
    
    % bundle all entities
    EE = vsa.bundleComplex(encEsin, encEcos, entity_weights);
    
    
end

% Preprocess the appearance features: projection to params.nDims,
% mean-centering, FFT and conversion to angle-representation.
% Slightly different to the paper, we use the pca-reduced descriptors
% (40-D) instead of the original (1024-D) features to save some web-traffic. 
% The performance is very similar.
%
% Y ... list of features
% 
% AF ... matrix of features, each row is a local feature descriptor
%
function AF = preprocessAppearanceFeatures(Y, params)
    persistent PP
    if isempty(PP)
        rng(102);
        PP = randn(size(Y.descriptors_pca,2), params.nDims, 'single');
        PP = orth(PP')';
    end
    D = Y.descriptors_pca * PP;
            
    % normalize per image
    mu = mean(D);
    D = D-mu;
            
    % convert to FHRR
    AF = angle(fft(D')');
end

% Encode the local features with their poses
function [F, n, LKP] = encodeLocalFeatures(vsa, keypoints, centroid, delfDf, pixelList, v, nxFB, nyFB, h, w)

    idxList = sub2ind([h,w], pixelList(:,2), pixelList(:,1));

    M = zeros(h,w,'uint8');
    M(idxList)=1;
    
    % check which keypoints are in this entity
    entityKP = [];
    for i=1:size(keypoints,1)

        % if this keypoint is in this entity
        x = keypoints(i,2)+1; % tested
        y = keypoints(i,1)+1;
        
        if y>h
            continue;
        end
        
        %idx = sub2ind([h,w], y, x);
        if M(y,x)>0
            %if ismember(idx, idxList)
            entityKP(end+1) = i;
        end
    end

    % collect encodings of all relevant keypoints
    LKP = zeros(numel(entityKP), vsa.d, 'single');
    idxLKP=1;
    for i=entityKP    
        x = keypoints(i,2)+1 - centroid(1); % tested
        y = keypoints(i,1)+1 - centroid(2);
        
        xPos = vsa.encode(v.baseX, x, 1/(w/nxFB));
        yPos = vsa.encode(v.baseY, y, 1/(h/nyFB));

        LKP(idxLKP,:) = vsa.bind([xPos;yPos;delfDf(i,:)]);
        idxLKP = idxLKP+1;
    end
    
    % bundle
    F = vsa.bundle(LKP);
    
    n = numel(entityKP);
end


% Weighted bundle of bind(orientation, neighboredClass)
function N = encodeGradStat(vsa, gradStat, v)    
    w = gradStat(:);
    N = vsa.bundleComplex(v.sinHistCellEnc, v.cosHistCellEnc, w);    
end


% Bundles all grid cells that are covered by the pixels in pixelList
function S = shapeFromSimpleGridEncoding(vsa, pixelList, v, nx, ny, h, w)

    xStep = w/nx;
    yStep = h/ny;

    xIdx = floor(pixelList(:,1)/xStep)+1; xIdx=min(xIdx,nx);
    yIdx = floor(pixelList(:,2)/yStep)+1; yIdx=min(yIdx,ny);
    
    % compute coverage for each grid cell
    GCC = zeros(ny, nx);
    for i=1:numel(xIdx)
        GCC( yIdx(i), xIdx(i) ) = GCC( yIdx(i), xIdx(i) ) + 1;
    end
    GCC = GCC/(xStep*yStep);
    
    % compute weight and bundle
    w = GCC(:);
    S = vsa.bundleComplex(v.gridSin, v.gridCos, w);
    
end
