% Compute the set of semantic entities with properties based on a semantic
% segmentation.
%
% L ... label image
% relevantClasses ... numeric class labels, have to be integers larger than 0
%
% E ... basically the output of regionsprops with additional class and
%       semantic gradient histogram
% classIdx{c} ... vector with all indices in E with class label c
%
function [E, classIdx] = getEntities(L, relevantClasses, params)
    
    assert(min(L(:))>0); % we can't handle class label 0 in classIdx

    E=[];

    % collect entities of each class
    for c = relevantClasses
        classIdxC = [];
        
        LC = L==c;

        stats = regionprops(LC, 'Area', 'Centroid', 'BoundingBox', 'Orientation','MajorAxisLength', 'MinorAxisLength', 'PixelList');
        [h,w] = size(L);
        for i=1:numel(stats)
            curStats = stats(i);
            
            % use only reasonably large objects
            if curStats.Area >= params.minEntSize
                
                % add class
                curStats.Class = c;
                
                % add semantic gradient histogram
                gradStat = getSemanticGradientHistogram(curStats.PixelList, L);
                curStats.GradStat = gradStat(:, relevantClasses);
                
                % add image size
                curStats.ImageSize = [h w];
                
                if isempty(E)
                    E = curStats;
                else
                    E(end+1) = curStats;
                end
                
                classIdxC(end+1) = numel(E);
            end
        end
        
        classIdx{c} = classIdxC;
    end

end


% Get the statistics of the spatial relations: How often is each class
% neighbored in each of the 8 cannoncical directions?
% Currently, this function is slow and not very powerful. Feel free to
% improve it!
%
% L ... must only contain labels 1:nClasses, nClasses is assumed to be 20
%       (currently this is designed for the Cityscapes labels:
%         1 road
%         2 sidewalk
%         3 building
%         4 wall
%         5 fence
%         6 pole
%         7 trafficlight
%         8 trafficsign
%         9 vegetation
%         10 terrain
%         11 sky
%         12 person
%         13 rider
%         14 car
%         15 truck
%         16 bus
%         17 train
%         18 motorcycle
%         19 bicycle
%         20 void)
%
% gradStat(direction,class) ... number of neighbors of a particular class in a
%                               particular direction (1:8)
%
function gradStat = getSemanticGradientHistogram(pixelList, L)
    
    assert(min(L(:))>0);
    assert(max(L(:))<=20);

    nClasses = 20;
    gradStat = zeros(8, nClasses);
    
    [h,w] = size(L);
    
    % ignore image boundary pixels
    idx = pixelList(:,1)==1 | pixelList(:,1)==w |  pixelList(:,2)==1 | pixelList(:,2)==h;
    validPixelList = pixelList;
    validPixelList(idx, :) = [];
    
    % for each valid pixel
    for i=1:size(validPixelList, 1)
       
        x = validPixelList(i,1);
        y = validPixelList(i,2);
        
        gradStat(1, L(y  ,x+1)) =  gradStat(1, L(y  ,x+1)) +1; % E
        gradStat(2, L(y+1,x+1)) =  gradStat(2, L(y+1,x+1)) +1; % SE
        gradStat(3, L(y+1,x  )) =  gradStat(3, L(y+1,x  )) +1; % S
        gradStat(4, L(y+1,x-1)) =  gradStat(4, L(y+1,x-1)) +1; % SW
        gradStat(5, L(y  ,x-1)) =  gradStat(5, L(y  ,x-1)) +1; % W
        gradStat(6, L(y-1,x-1)) =  gradStat(6, L(y-1,x-1)) +1; % NW
        gradStat(7, L(y-1,x  )) =  gradStat(7, L(y-1,x  )) +1; % N
        gradStat(8, L(y-1,x+1)) =  gradStat(8, L(y-1,x+1)) +1; % NE
        
    end

end


